/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Koetter interpolation algorithm. (single multiplicity variant)
   From: The Guruswami-Sudan Decoding Algorithm for Reed-Solomon Codes.
         R. J. McEliece. */

/**
> `int ll_koetter_single(R *q,int *nqx,int *nqy,R *supp,int n,int k,\
  int L,int m,R *y,int tau);`
  Set `q` of size (`nqx`,`nqy`) to be the interpolation polynomial of
  the Guruswami-Sudan list decoding algorithm for a Reed-Solomon code
  of support `supp`, length `n`, dimension `k`, list-size `L` and for
  a received word `y` with at most `tau` errors. The multiplicity `m`
  is the same for all points (`supp[i]`,`y[i]`).
  The Koetter algorithm is used.
*/

int ll_koetter_single(R *q,int *nqx,int *nqy,
                      R *supp,int n,int k,
                      int L,int m,R *y,int tau)
{
  R **g,x0,y0,*f,delta,*bufR,*t,coeff,*xchg;
  int *wdeg,*ngy,*mark,*bufint;
  int ngx,ngy_max,i,j,r,s,j_star,ngy_star,wdeg_star,L1;
  int wdeg_max,index,ret;
  int ng_max,nbufR,pc_ngy;
#ifdef RING_COMM_CAR2
  struct ll_dbpol_shift_car2_pc *pc;
#else
  struct ll_dbpol_shift_fast_pc *pc;
#endif

  /* to make gcc happy */
  f = NULL;

  /* default ret code */
  ret = 0;
  
  /* malloc int's */
  L1 = L + 1;
  bufint = ll_malloc(3 * L1 * sizeof(int));
  wdeg = bufint;
  ngy = &bufint[L1];
  mark = &bufint[2 * L1];
  
  /* malloc R's */
  ngx = L1;
  ngy_max = m * (n - tau) + 1;
  ng_max = ngx * ngy_max;
  nbufR = (L1 + 1) * ng_max;
  bufR = ll_vec_malloc_init(nbufR);
  t = &bufR[L1 * ng_max];
  g = ll_malloc(L1 * sizeof(R*));
  
  /* init other stuff */
  for( j = 0 ; j <= L ; j++ ) {
    g[j] = &bufR[j * ng_max];
    mark[j] = 0;
    ngy[j] = 1;
    wdeg[j] = (k - 1) * j;
    R_castint(g[j][j],1);
  }
  wdeg_max = (ngy_max - 1) + (k - 1) * (ngx - 1);
  pc_ngy = 0;
  
  /* go for it */
  R_init(x0);
  R_init(y0);
  R_init(delta);
  R_init(coeff);
  for( i = 0 ; i < n ; i++ ) {

    /* shift to obtain all the D_{rs} */
    R_sub(x0,supp[i],x0);
    R_sub(y0,y[i],y0);
#ifdef RING_COMM_CAR2
    pc = ll_dbpol_shift_car2_prepare(ngx,pc_ngy,y0,x0);
    for( j = 0 ; j <= L ; j++ ) {
      if ( mark[j] < 2 ) {
        ll_dbpol_shift_car22(t,g[j],ngx,ngy[j],y0,x0,pc);
        xchg = t;
        t = g[j];
        g[j] = xchg;
      }
    }
    ll_dbpol_shift_car2_finish(pc);
#else
    pc = ll_dbpol_shift_fast_prepare(ngx,pc_ngy,y0,x0);
    for( j = 0 ; j <= L ; j++ ) {
      if ( mark[j] < 2 ) {
        ll_dbpol_shift_fast2(t,g[j],ngx,ngy[j],y0,x0,pc);
        xchg = t;
        t = g[j];
        g[j] = xchg;
      }
    }
    ll_dbpol_shift_fast_finish(pc);
#endif
    R_copy(x0,supp[i]);
    R_copy(y0,y[i]);
   
    /* main loop */
    for( r = 0 ; r < m ; r++ ) {
      for( s = 0 ; s < m - r ; s++ ) {
        
        /* find the smallest of the gj's */
        j_star = -1;
        ngy_star = ngy_max + 1;
        wdeg_star = wdeg_max + 1;
        index = ngx * r + s;
        for( j = 0 ; j <= L ; j++ ) {
          if ( mark[j] < 2 ) {
            if ( !R_iszero(g[j][index]) ) {
              mark[j] = 1;
              if (   wdeg[j] <  wdeg_star ||
                   ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
              {
                f = g[j];
                j_star = j;
                ngy_star = ngy[j];
                wdeg_star = wdeg[j];
              }
            }
            else {
              mark[j] = 0;
            }
          }
        }
        
        /* update the gj's */
        if ( j_star > -1 ) {
          if ( R_inv(delta,f[index]) ) {
            ll_die("Koetter(): delta is not invertible");
          }
          R_neg(delta,delta);
          for( j = 0 ; j <= L ; j++ ) {
            if ( mark[j] == 1 ) {
              if ( j == j_star ) {
                if ( ngy[j] >= ngy_max ) {
                  mark[j] = 2;
                }
                else {
                  ll_dbpol_mul_by_Y(t,f,ngx,ngy_star);
                }
              }
              else {
                R_mul(coeff,delta,g[j][index]);
                ll_array_add_mul_by_cte(g[j],ngx * ngy[j],
                                        f,ngx * ngy_star,coeff);
                ngy[j] = ll_max(ngy[j],ngy_star);
              }
            }
          }
          xchg = t;
          t = g[j_star];
          g[j_star] = xchg;
          ngy[j_star]++;
          if ( ngy[j_star] > pc_ngy ) pc_ngy = ngy[j_star];
          wdeg[j_star]++;
        }
        else {
          ret = -1;
          goto koetter_free_stuff;
        }
      
      }
    }
  
  }
  
  /* select the min of the gj's */
  j_star = -1;
  ngy_star = ngy_max + 1;
  wdeg_star = wdeg_max + 1;
  for( j = 0 ; j <= L ; j++ ) {
    if ( mark[j] < 2 ) {
      if ( wdeg[j] <  wdeg_star ||
         ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
      {
        f = g[j];
        j_star = j;
        ngy_star = ngy[j];
        wdeg_star = wdeg[j];
      }
    }
  }
  if ( j_star == -1 ) {
    ret = -1;
    goto koetter_free_stuff;
  }
  
  /* shift back the candidate */
  R_neg(x0,supp[n - 1]);
  R_neg(y0,y[n - 1]);
  ll_dbpol_shift(q,f,ngx,ngy_star,y0,x0);
  nqx[0] = ngx;
  nqy[0] = ngy_star;

koetter_free_stuff:
  ll_free(bufint);
  ll_free(g);
  ll_vec_free(bufR,nbufR);
  R_clear(coeff);
  R_clear(delta);
  R_clear(y0);
  R_clear(x0);
  
  return ret;
}

/**
> `int ll_koetter_multi(R *q,int *nqx,int *nqy,R *supp,int n,int k,\
  int L,int *m,R *y,int tau);`
  Set `q` of size (`nqx`,`nqy`) to be the interpolation polynomial of
  the Guruswami-Sudan list decoding algorithm for a Reed-Solomon code
  of support `supp`, length `n`, dimension `k`, list-size `L` and for
  a received word `y` with at most `tau` errors. The multiplicities
  for each point (`supp[i]`,`y[i]`) is given by `m[i]`.
  The Koetter algorithm is used.
*/

int ll_koetter_multi(R *q,int *nqx,int *nqy,
                     R *supp,int n,int k,
                     int L,int *m,R *y,int tau)
{
  R **g,x0,y0,*f,delta,*bufR,*t,coeff,*xchg;
  int *wdeg,*ngy,*mark,*bufint;
  int ngx,ngy_max,i,j,r,s,j_star,ngy_star,wdeg_star,L1;
  int wdeg_max,index,ret;
  int ng_max,nbufR,pc_ngy,m_max;
#ifdef RING_COMM_CAR2
  struct ll_dbpol_shift_car2_pc *pc;
#else
  struct ll_dbpol_shift_fast_pc *pc;
#endif

  /* to make gcc happy */
  f = NULL;

  /* default ret code */
  ret = 0;
  
  /* malloc int's */
  L1 = L + 1;
  bufint = ll_malloc(3 * L1 * sizeof(int));
  wdeg = bufint;
  ngy = &bufint[L1];
  mark = &bufint[2 * L1];
  
  /* malloc R's */
  m_max = m[0];
  for( i = 1 ; i < n ; i++ ) {
    if ( m[i] > m_max ) m_max = m[i];
  }
  ngx = L1;
  ngy_max = m_max * (n - tau) + 1;
  ng_max = ngx * ngy_max;
  nbufR = (L1 + 1) * ng_max;
  bufR = ll_vec_malloc_init(nbufR);
  t = &bufR[L1 * ng_max];
  g = ll_malloc(L1 * sizeof(R*));
  
  /* init other stuff */
  for( j = 0 ; j <= L ; j++ ) {
    g[j] = &bufR[j * ng_max];
    mark[j] = 0;
    ngy[j] = 1;
    wdeg[j] = (k - 1) * j;
    R_castint(g[j][j],1);
  }
  wdeg_max = (ngy_max - 1) + (k - 1) * (ngx - 1);
  pc_ngy = 0;
  
  /* go for it */
  R_init(x0);
  R_init(y0);
  R_init(delta);
  R_init(coeff);
  for( i = 0 ; i < n ; i++ ) {

    /* shift to obtain all the D_{rs} */
    R_sub(x0,supp[i],x0);
    R_sub(y0,y[i],y0);
#ifdef RING_COMM_CAR2
    pc = ll_dbpol_shift_car2_prepare(ngx,pc_ngy,y0,x0);
    for( j = 0 ; j <= L ; j++ ) {
      if ( mark[j] < 2 ) {
        ll_dbpol_shift_car22(t,g[j],ngx,ngy[j],y0,x0,pc);
        xchg = t;
        t = g[j];
        g[j] = xchg;
      }
    }
    ll_dbpol_shift_car2_finish(pc);
#else
    pc = ll_dbpol_shift_fast_prepare(ngx,pc_ngy,y0,x0);
    for( j = 0 ; j <= L ; j++ ) {
      if ( mark[j] < 2 ) {
        ll_dbpol_shift_fast2(t,g[j],ngx,ngy[j],y0,x0,pc);
        xchg = t;
        t = g[j];
        g[j] = xchg;
      }
    }
    ll_dbpol_shift_fast_finish(pc);
#endif
    R_copy(x0,supp[i]);
    R_copy(y0,y[i]);
   
    /* main loop */
    for( r = 0 ; r < m[i] ; r++ ) {
      for( s = 0 ; s < m[i] - r ; s++ ) {
        
        /* find the smallest of the gj's */
        j_star = -1;
        ngy_star = ngy_max + 1;
        wdeg_star = wdeg_max + 1;
        index = ngx * r + s;
        for( j = 0 ; j <= L ; j++ ) {
          if ( mark[j] < 2 ) {
            if ( !R_iszero(g[j][index]) ) {
              mark[j] = 1;
              if (   wdeg[j] <  wdeg_star ||
                   ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
              {
                f = g[j];
                j_star = j;
                ngy_star = ngy[j];
                wdeg_star = wdeg[j];
              }
            }
            else {
              mark[j] = 0;
            }
          }
        }
        
        /* update the gj's */
        if ( j_star > -1 ) {
          if ( R_inv(delta,f[index]) ) {
            ll_die("Koetter(): delta is not invertible");
          }
          R_neg(delta,delta);
          for( j = 0 ; j <= L ; j++ ) {
            if ( mark[j] == 1 ) {
              if ( j == j_star ) {
                if ( ngy[j] >= ngy_max ) {
                  mark[j] = 2;
                }
                else {
                  ll_dbpol_mul_by_Y(t,f,ngx,ngy_star);
                }
              }
              else {
                R_mul(coeff,delta,g[j][index]);
                ll_array_add_mul_by_cte(g[j],ngx * ngy[j],
                                        f,ngx * ngy_star,coeff);
                ngy[j] = ll_max(ngy[j],ngy_star);
              }
            }
          }
          xchg = t;
          t = g[j_star];
          g[j_star] = xchg;
          ngy[j_star]++;
          if ( ngy[j_star] > pc_ngy ) pc_ngy = ngy[j_star];
          wdeg[j_star]++;
        }
        else {
          ret = -1;
          goto koetter_free_stuff;
        }
      
      }
    }
  
  }
  
  /* select the min of the gj's */
  j_star = -1;
  ngy_star = ngy_max + 1;
  wdeg_star = wdeg_max + 1;
  for( j = 0 ; j <= L ; j++ ) {
    if ( mark[j] < 2 ) {
      if ( wdeg[j] <  wdeg_star ||
         ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
      {
        f = g[j];
        j_star = j;
        ngy_star = ngy[j];
        wdeg_star = wdeg[j];
      }
    }
  }
  if ( j_star == -1 ) {
    ret = -1;
    goto koetter_free_stuff;
  }
  
  /* shift back the candidate */
  R_neg(x0,supp[n - 1]);
  R_neg(y0,y[n - 1]);
  ll_dbpol_shift(q,f,ngx,ngy_star,y0,x0);
  nqx[0] = ngx;
  nqy[0] = ngy_star;

koetter_free_stuff:
  ll_free(bufint);
  ll_free(g);
  ll_vec_free(bufR,nbufR);
  R_clear(coeff);
  R_clear(delta);
  R_clear(y0);
  R_clear(x0);
  
  return ret;
}


