/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <decoding/ring_reset.h>

#define R GF64
#define GF_MAGMA_DECL "k := GF(64)"

#define RING_COMM_CAR2

/* defining polynomial is x^6 + x + 1 */

typedef int R[1];

void R_print(FILE *out,R a) {
  fputc('(',out);
  fputc(((a[0] & (1 << 5)) ? '1' : '0'),out);
  fputc(((a[0] & (1 << 4)) ? '1' : '0'),out);
  fputc(((a[0] & (1 << 3)) ? '1' : '0'),out);
  fputc(' ',out);
  fputc(((a[0] & (1 << 2)) ? '1' : '0'),out);
  fputc(((a[0] & (1 << 1)) ? '1' : '0'),out);
  fputc(((a[0] & (1     )) ? '1' : '0'),out);
  fputc(')',out);
}

int R_reduce(int a) {
  int b;
  b = a & 63;
  b ^= (-((a >> 6 ) & 1)) &  3;
  b ^= (-((a >> 7 ) & 1)) & (3 << 1);
  b ^= (-((a >> 8 ) & 1)) & (3 << 2);
  b ^= (-((a >> 9 ) & 1)) & (3 << 3);
  b ^= (-((a >> 10) & 1)) & (3 << 4);
  return b;
}

void R_mul(R c,R a,R b) {
  int ab;
  ab  = (-( b[0]       & 1)) &  a[0];
  ab ^= (-((b[0] >> 1) & 1)) & (a[0] << 1);
  ab ^= (-((b[0] >> 2) & 1)) & (a[0] << 2);
  ab ^= (-((b[0] >> 3) & 1)) & (a[0] << 3);
  ab ^= (-((b[0] >> 4) & 1)) & (a[0] << 4);
  ab ^= (-((b[0] >> 5) & 1)) & (a[0] << 5);
  c[0] = R_reduce(ab);
}

int R_deg(int a) {
  if ( a & (int)(1 << 5) ) return 5;
  if ( a & (int)(1 << 4) ) return 4;
  if ( a & (int)(1 << 3) ) return 3;
  if ( a & (int)(1 << 2) ) return 2;
  if ( a & (int)(1 << 1) ) return 1;
  return 0;
}

int R_inv(R c,R a) {
  int dr0, dr1, dr2;
  int b,r0,r1,r2,v0,v1,v2,qv1,m;
  if ( a[0] == 0 ) return -1;
  b = (1 << 6) ^ (1 << 1) ^ 1;
  r0 = b; dr0 = 6;
  r1 = a[0]; dr1 = R_deg(r1);
  v0 = 0;
  v1 = 1;
  for( r2 = 1 ; r2 ; ) {
    r2 = r0; dr2 = dr0;
    qv1 = 0;
    m = 1 << dr2;
    while( dr2 >= dr1 ) {
      int tmp;
      tmp = dr2 - dr1;
      r2  ^= r1 << tmp;
      qv1 ^= v1 << tmp;
      for( m >>= 1,dr2-- ; !(m & r2) && dr2 >= 0; dr2--,m >>= 1 );
    }
    v2= v0 ^ qv1;
    v0= v1; v1= v2;
    r0= r1; r1= r2;
    dr0= dr1; dr1= dr2;
  }
  c[0] = v0;
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0;
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0;
  for(;;) {
    int *b;
    b = va_arg(ap,int*);
    if ( !b ) return;
    b[0] = 0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0);
}

void R_zero(R r) {
  r[0] = 0;
}

void R_random(R r) {
  r[0] = rand() & 63;
}

void R_add(R r,R a,R b) {
  r[0] = a[0] ^ b[0];
}

void R_sub(R r,R a,R b) {
  r[0] = a[0] ^ b[0];
}

void R_sub_sub(R r,R a,R b) {
  r[0] = r[0] ^ a[0] ^ b[0];
}

void R_add_mul(R r,R a,R b) {
  R c;
  R_mul(c,a,b);
  r[0] = r[0] ^ c[0];
}

void R_sub_mul(R r,R a,R b) {
  R c;
  R_mul(c,a,b);
  r[0] = r[0] ^ c[0];
}

void R_neg(R r,R a) {
  r[0] = a[0];
}

void R_square(R r,R a) {
  R_mul(r,a,a);
}

void R_castint(R r,int a) {
  if ( a >= 0 ) r[0] = a & 1; else r[0] = (-a) & 1;
}

void R_next(R r,R a) {
  r[0] = a[0] + 1;
  if ( r[0] == 64 ) r[0] = 0;
}

#include <decoding/algos.h>

#define MY_UPOL_ROOTS GF64_my_upol_roots
void MY_UPOL_ROOTS(R *r,int *nr,R *a,int na) {
  int i;
  R x,y;
  nr[0] = 0;
  for( i = 0 ; i < 64 ; i++ ) {
    x[0] = i;
    ll_upol_horner(y,a,na,x);
    if ( y[0] == 0 ) {
      R_copy(r[nr[0]],x);
      nr[0]++;
    }
  }
}

