/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `gfp_word.c`
  Implement all prime fields (`Z/pZ`) whose characteristic holds
  within a machine word. Use `PRIME` to indicate the mudulus of
  the finite fields.

  ``
  #define PRIME 101
  #include <decoding/rings/gfp_word.h>
  #include <decoding/algos.c>
  ``
  
  In the example above, the name of the finite field will be
  `gf101`. You can of course use several prime fields in the
  same program.

  ``
  #define PRIME 7
  #include <decoding/rings/gfp_word.h>
  #include <decoding/algos.c>
  
  #define PRIME 11
  #include <decoding/rings/gfp_word.h>
  #include <decoding/algos.c>
  ``

  It will then generate the code for the two finite fields
  GF(7) and GF(11). You will then have two types.

  - `gf7` corresponding to GF(7).

  - `gf11` corresponding to GF(11).
*/

#include <decoding/ring_reset.h>

#define R __C(gf,PRIME)
#define GF_MAGMA_DECL "k := GF()"

typedef int R[1];

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0;
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0;
  for(;;) {
    int *b;
    b = va_arg(ap,int*);
    if ( !b ) return;
    b[0] = 0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

void R_print(FILE *out,R a) {
  fprintf(out,"%d",a[0]);
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0);
}

void R_zero(R r) {
  r[0] = 0;
}

void R_random(R r) {
  r[0] = rand() % PRIME;
}

void R_add(R r,R a,R b) {
  r[0] = (a[0] + b[0]) % PRIME;
}

void R_sub(R r,R a,R b) {
  r[0] = (a[0] + 17 - b[0]) % PRIME;
}

void R_sub_sub(R r,R a,R b) {
  r[0] = (r[0] + PRIME + PRIME - a[0] - b[0]) % PRIME;
}

void R_add_mul(R r,R a,R b) {
  r[0] = (r[0] + (a[0] * b[0])) % PRIME;
}

void R_sub_mul(R r,R a,R b) {
  r[0] = (r[0] + (PRIME * PRIME) - (a[0] * b[0])) % PRIME;
}

void R_neg(R r,R a) {
  r[0] = PRIME - a[0];
}

void R_mul(R r,R a,R b) {
  r[0] = (a[0] * b[0]) % PRIME;
}

void R_square(R r,R a) {
  r[0] = (a[0] * a[0]) % PRIME;
}

void R_castint(R r,int a) {
  if ( a >= 0 ) {
    r[0] = a % PRIME;
  }
  else {
    r[0] = PRIME - ((-a) % PRIME);
  }
}

int R_inv(R r,R a) {
  int q,r0,r1,r2,s0,s1,s2,t0,t1,t2;
  if ( a[0] == 0 ) return -1;
  r0 = PRIME; r1 = a[0];
  s0 = 1;     s1 = 0;
  t0 = 0;     t1 = 1;
  for( ; r1 != 0 ; ) {
    q = r0 / r1;
    r2 = r0 - q * r1;
    s2 = s0 - q * s1;
    t2 = t0 - q * t1;
    r0 = r1; r1 = r2;
    s0 = s1; s1 = s2;
    t0 = t1; t1 = t2;
  }
  R_castint(r,t0);
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_next(R r,R a) {
  r[0] = a[0] + 1;
  if ( r[0] == PRIME ) r[0] = 0;
}

#include <decoding/algos.h>

