/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <decoding/rings/GF2.c>
/* #include "../rings/GF64.c" */
#include <decoding/algos.c>

void mtest_upol_horner(FILE *out,int n1,int n2) {
  int i;
  R *f,x,y;
  f = ll_upol_malloc_random(n2);
  R_init(x);
  R_init(y);
  for( i = n1 ; i <= n2 ; i++ ) {
    R_random(x);
    ll_upol_horner(y,f,i,x);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  x  := k ! (%e);\n"
                 "  y  := k ! (%e);\n"
                 "  if not (Evaluate(f,x) eq y) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,&x,&y);
  }
  ll_vec_free(f,n2);
  R_clear(x);
  R_clear(y);
}

void mtest_upol_mul(FILE *out,int n1,int n2) {
  int i,j,m;
  R *f,*g,*fg;
  m = 2 * n2 - 1;
  f = ll_upol_malloc_random(n2);
  g = ll_upol_malloc_random(n2);
  fg = ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    ll_upol_mul(fg,f,i,g,j);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  g  := P ! (%p);\n"
                 "  fg := P ! (%p);\n"
                 "  if not (fg eq (f * g)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,fg,i + j - 1);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  ll_vec_free(fg,m);
}

void mtest_upol_square(FILE *out,int n1,int n2) {
  int i,m;
  R *f,*f2;
  m = 2 * n2 - 1;
  f = ll_upol_malloc_random(n2);
  f2 = ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    ll_upol_square(f2,f,i);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  f2 := P ! (%p);\n"
                 "  if not (f2 eq (f * f)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,f2,2 * i - 1);
  }
  ll_vec_free(f,n2);
  ll_vec_free(f2,m);
}
void mtest_upol_eucldiv_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  R *f,*g,*r,*q;
  f = ll_vec_malloc_init(n2);
  g = ll_vec_malloc_init(n2);
  r = ll_vec_malloc_init(n2);
  q = ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    ll_upol_random(f,i);
    ll_upol_random(g,j);
    ll_upol_quorem_naive(r,&nr,q,f,i,g,j);
    ll_print(out,"if e eq 0 then\n"
                 "  f := P ! (%p);\n"
                 "  g := P ! (%p);\n"
                 "  r := P ! (%p);\n"
                 "  q := P ! (%p);\n"
                 "  if not ( (q eq (f div g)) and\n"
                 "           (r eq (f mod g)) and\n"
                 "           (Degree(r) eq (%d - 1)) ) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  ll_vec_free(r,n2);
  ll_vec_free(q,n2);
}

void mtest_upol_eucldiv2_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  R *f,*g,*r,*q;
  f = ll_vec_malloc_init(n2);
  g = ll_vec_malloc_init(n2);
  r = ll_vec_malloc_init(n2);
  q = ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    ll_upol_random(f,i);
    ll_upol_random(g,j);
    ll_upol_quo_naive(q,f,i,g,j);
    ll_upol_rem_naive(r,&nr,f,i,g,j);
    ll_print(out,"if e eq 0 then\n"
                 "  f := P ! (%p);\n"
                 "  g := P ! (%p);\n"
                 "  r := P ! (%p);\n"
                 "  q := P ! (%p);\n"
                 "  if not ( (q eq (f div g)) and\n"
                 "           (r eq (f mod g)) and\n"
                 "           (Degree(r) eq (%d - 1)) ) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  ll_vec_free(r,n2);
  ll_vec_free(q,n2);
}

void mtest_upol_gcd_naive(FILE *out,int n1,int n2) {
  int i,j,ngcd;
  R *f,*g,*gcd;
  f = ll_vec_malloc_init(n2);
  g = ll_vec_malloc_init(n2);
  gcd = ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    ll_upol_random(f,i);
    ll_upol_random(g,j);
    ll_upol_gcd_naive(gcd,&ngcd,f,i,g,j);
    ll_upol_normalize(gcd,ngcd);
    ll_print(out,"if e eq 0 then\n"
                 "  f   := P ! (%p);\n"
                 "  g   := P ! (%p);\n"
                 "  gcd := P ! (%p);\n"
                 "  if not (gcd eq Gcd(f,g)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,gcd,ngcd);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  ll_vec_free(gcd,n2);
}

void mtest_upol_shift_naive(FILE *out,int n1,int n2) {
  int i;
  R *f,*g,b;
  f = ll_upol_malloc_random(n2);
  g = ll_vec_malloc_init(n2);
  R_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    R_random(b);
    ll_upol_shift_naive(g,f,i,b);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,&b,g,i);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  R_clear(b);
}

void mtest_log2(FILE *out,int n1,int n2) {
  int i;
  for( i = n1 ; i <= n2 ; i++ ) {
    ll_print(out,"if e eq 0 then\n"
                 "  n := %d;\n"
                 "  l := %d;\n"
                 "  if not (l eq Ilog2(n)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",i,ll_log2(i));
  }
}

void mtest_prev_log2(FILE *out,int n1,int n2) {
  int i;
  for( i = n1 ; i <= n2 ; i++ ) {
    ll_print(out,"if e eq 0 then\n"
                 "  n := %d;\n"
                 "  l := %d;\n"
                 "  if (2^l ge n) or (l lt Ilog2(n) - 1) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",i,ll_prev_log2(i));
  }
}

void mtest_upol_shift_fast(FILE *out,int n1,int n2) {
  int i;
  R *f,*g,b;
  f = ll_upol_malloc_random(n2);
  g = ll_vec_malloc_init(n2);
  R_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    R_random(b);
    ll_upol_shift_fast(g,f,i,b);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,&b,g,i);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  R_clear(b);
}

void mtest_upol_shift_car2(FILE *out,int n1,int n2) {
  int i;
  R *f,*g,b;
  f = ll_upol_malloc_random(n2);
  g = ll_vec_malloc_init(n2);
  R_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    R_random(b);
    ll_upol_shift_car2(g,f,i,b);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,&b,g,i);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  R_clear(b);
}

void mtest_upol_shift_car2_inplace(FILE *out,int n1,int n2) {
  int i;
  R *f,*g,b;
  f = ll_upol_malloc_random(n2);
  g = ll_vec_malloc_init(n2);
  R_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    R_random(b);
    ll_vec_copy(g,f,i);
    ll_upol_shift_car2(g,g,i,b);
    ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,&b,g,i);
  }
  ll_vec_free(f,n2);
  ll_vec_free(g,n2);
  R_clear(b);
}

#define D1 5
#define D2 100

int main(void) {
  srand(time(0));
  ll_print(stdout,"clear;\n"
                   GF_MAGMA_DECL ";\n"
                  "P<X> := PolynomialRing(k);\n"
                  "e := 0;\n\n");
  mtest_upol_horner(stdout,D1,D2);
  mtest_upol_mul(stdout,D1,D2);
  mtest_upol_square(stdout,D1,D2);
  mtest_upol_eucldiv_naive(stdout,D1,D2);
  mtest_upol_eucldiv2_naive(stdout,D1,D2);
  mtest_upol_gcd_naive(stdout,D1,D2);
  mtest_upol_shift_naive(stdout,D1,D2);
  mtest_log2(stdout,D1,D2);
  mtest_prev_log2(stdout,D1,D2);
  mtest_upol_shift_fast(stdout,D1,D2);
  mtest_upol_shift_car2(stdout,D1,D2);
  mtest_upol_shift_car2_inplace(stdout,D1,D2);
  ll_print(stdout,"if e eq 0 then\n"
                  "  print \"ALL OK\";\n"
                  "end if;\n\n");
  return 0;
}

